import pytest
import numpy as np
from openairbearing.plots import plot_key_results, plot_bearing_shape
from openairbearing.bearings import (
    BaseBearing,
    CircularBearing,
    AnnularBearing,
    InfiniteLinearBearing,
    RectangularBearing,
)
from plotly.graph_objects import Figure


# def test_plot_key_results():
#     """Test the plot_key_results function."""
#     # Mock a bearing object
#     bearings = [
#         CircularBearing(),
#         AnnularBearing(),
#         InfiniteLinearBearing(),
#         RectangularBearing(),
#     ]
#     for bearing in bearings:

#         # Mock results
#         class MockResult:
#             def __init__(self, name):
#                 self.name = name
#                 self.w = np.random.rand(bearing.nh)
#                 self.k = np.random.rand(bearing.nh)
#                 self.qs = np.random.rand(bearing.nh)
#                 self.qc = np.random.rand(bearing.nh)
#                 self.qa = np.random.rand(bearing.nh)
#                 if bearing.case == "rectangular":
#                     self.p = np.random.rand(bearing.ny, bearing.nx, bearing.nh)
#                 else:
#                     self.p = np.random.rand(bearing.nx, bearing.nh)

#         results = [MockResult("Analytic"), MockResult("Numeric")]

#         # Call the function
#         fig = plot_key_results(bearing, results)

#         # Assertions
#         assert isinstance(fig, Figure), "The output should be a Plotly Figure."
#         assert len(fig.data) > 0, "The figure should contain traces."
#         assert len(fig.layout.annotations) > 0, "The figure should have subplot titles."


# def test_plot_key_results_no_results():
#     """Test plot_key_results with no results."""
#     # Mock a bearing object
#     bearing = CircularBearing()

#     # Call the function with no results
#     fig = plot_key_results(bearing, [])

#     # Assertions
#     assert isinstance(fig, Figure), "The output should be a Plotly Figure."
#     assert len(fig.data) == 0, "The figure should contain no traces."
#     assert (
#         fig.layout.title.text == "No suitable solver selected"
#     ), "The figure should display a message when no results are provided."


# def test_plot_bearing_shape():
#     """Test the plot_bearing_shape function."""
#     # Mock a bearing object
#     bearings = [
#         CircularBearing(),
#         AnnularBearing(),
#         InfiniteLinearBearing(),
#         RectangularBearing(),
#     ]
#     for bearing in bearings:
#         bearing.ha = np.linspace(bearing.ha_min, bearing.ha_max, bearing.nx)
#         bearing.x = np.linspace(0, bearing.xa, bearing.nx)

#         # Call the function
#         fig = plot_bearing_shape(bearing)

#         # Assertions
#         assert isinstance(fig, Figure), "The output should be a Plotly Figure."
#         assert len(fig.data) > 0, "The figure should contain traces."
#         assert len(fig.layout.annotations) > 0, "The figure should have subplot titles."



import numpy as np
from dataclasses import dataclass, field

from openairbearing.utils import get_area, get_geom, get_kappa, get_beta


@dataclass
class BaseBearing:
    """Base class for all bearing types."""

    pa: float = 101325
    pc: float = pa
    ps: float = 0.6e6 + pa

    rho: float = 1.293
    mu: float = 1.85e-5

    hp: float = 4.5e-3

    ha_min: float = 1e-6
    ha_max: float = 20e-6

    xa: float = 37 / 2 * 1e-3
    xc: float = 0
    ya: float = 0
    nh: int = 20
    nx: int = 30
    ny: int = 1

    Psi: float = 0

    error_type: str = "none"
    error: float = 0e-6

    blocked: bool = False
    block_x: float = 25.2e-3 / 2
    block_w: float = 1e-3

    Qsc: float = 3  # L/min
    psc: float = 0.6e6 + pa

    # clearance and eccentricity for journal bearings
    c: float = field(init=False)
    e: float = field(init=False)
    theta: np.ndarray = field(init=False, default=None)
    clearance: np.ndarray = field(init=False, default=None)

    x: np.ndarray = field(init=False)
    dx: np.ndarray = field(init=False)
    y: np.ndarray = field(init=False)
    dy: np.ndarray = field(init=False)

    ha: np.ndarray = field(init=False)
    A: float = field(init=False)
    kappa: float = field(init=False)
    beta: float = field(init=False)
    geom: np.ndarray = field(init=False)

    case: str = "base"
    type: str = "bearing"
    csys: str = "cartesian"

    def __post_init__(self):
        self.ha = np.linspace(self.ha_min, self.ha_max, self.nh).T
        self.x = np.linspace(self.xc, self.xa, self.nx)
        self.y = np.linspace(0, self.ya, self.ny)
        self.dx = np.gradient(self.x)
        self.dy = 1 if self.ny == 1 else np.gradient(self.y)
        self.A = get_area(self)
        self.geom = get_geom(self)
        self.kappa = get_kappa(self)
        self.beta = get_beta(self)


@dataclass
class CircularBearing(BaseBearing):
    """Base class for circular thrust bearing"""

    case: str = "circular"
    type: str = "bearing"
    csys: str = "polar"

    xc: float = 1e-6
    xa: float = 37e-3 / 2
    Qsc: float = 2.8  # L/min

    def __post_init__(self):
        super().__post_init__()
        self.psc = 0.6e6 + self.pa


@dataclass
class AnnularBearing(BaseBearing):
    """Base class for annular bearing"""

    case: str = "annular"
    type: str = "seal"
    csys: str = "polar"

    xa: float = 58e-3 / 2
    xc: float = 25e-3 / 2

    Qsc: float = 3  # L/min

    def __post_init__(self):
        super().__post_init__()
        self.psc = 0.6e6 + self.pa


@dataclass
class InfiniteLinearBearing(BaseBearing):
    """Base class for Infinitely long linear bearing bearing"""

    case: str = "infinite"
    type: str = "seal"
    csys: str = "cartesian"

    ps: float = 0.41e6
    xa: float = 40e-3

    Qsc: float = 37  # L/min

    def __post_init__(self):
        super().__post_init__()
        self.psc = 0.41e6 + self.pa


@dataclass
class RectangularBearing(BaseBearing):
    """Base class for rectangular thrust bearing"""

    case: str = "rectangular"
    type: str = "bearing"
    csys: str = "cartesian"

    xa: float = 80e-3
    ya: float = 40e-3
    nx: int = 40
    ny: int = 20

    ps: float = 0.41e6

    Qsc: float = 2.94  # L/min

    def __post_init__(self):
        super().__post_init__()
        self.psc = 0.41e6 + self.pa
        self.x = np.linspace(-self.xa / 2, self.xa / 2, self.nx)
        self.y = np.linspace(-self.ya / 2, self.ya / 2, self.ny)
        self.dx = self.xa / (self.nx + 1)
        self.dy = self.ya / (self.ny + 1)
        self.geom = get_geom(self)  # calculate after x y


@dataclass
class JournalBearing(BaseBearing):
    """Base class for journal bearing"""

    case: str = "journal"
    type: str = "seal"
    csys: str = "cartesian"

    xa: float = 50.02e-3 / 2  # radius
    ya: float = 89e-3  # length
    nx: int = 80
    ny: int = 50
    hp: float = 3e-3

    c = 40e-6  # clearance (journal radius - shaft radius)

    ps: float = 0.41e6

    Qsc: float = 15  # L/min

    def __post_init__(self):
        super().__post_init__()
        self.psc = 0.41e6 + self.pa
        self.theta = np.linspace(-np.pi, np.pi, self.nx)
        self.x = self.theta
        self.y = np.linspace(-self.ya / 2, self.ya / 2, self.ny)
        self.dx = 2 * np.pi / (self.nx + 1)
        self.dy = self.ya / (self.ny)

        self.ha_min = 0.01e-6
        self.ha_max = self.c / 2 - 1e-6
        e_min = self.ha_min
        e_max = self.ha_max
        self.e = np.linspace(e_min, e_max, self.nh)
        self.clearance = self.xa - np.sqrt(
            (self.xa - self.c / 2) ** 2
            + self.e[None, None, :] ** 2
            + 2
            * self.e[None, None, :]
            * (self.xa - self.c / 2)
            * np.cos(self.theta[:, None, None])
        )
        # import matplotlib.pyplot as plt
        # plt.plot(self.theta, np.squeeze(self.clearance) * 1e6)
        # plt.plot(self.theta, np.cos(self.theta))
        # plt.show()
        # print(self.dx*self.nx)
        self.geom = get_geom(self)  # calculate after x y



"""OpenAir - Air Bearing Analysis Tool.

This package provides tools for analyzing and visualizing air bearing performance.
"""

# Version information
__version__ = "0.1.3"

# Import constants from config (where they're actually defined)
from .config import DEMO_MODE

# Import specific bearing classes
from .bearings import (
    RectangularBearing,
    CircularBearing,
    AnnularBearing,
    InfiniteLinearBearing,
    JournalBearing,
)

# Import utility functions from where they're defined
from .utils import (
    get_kappa,
    get_Qsc,
    get_beta,
    Result,
    get_geom,
    get_area,
)

# Import solver function
from .solvers import solve_bearing, get_pressure_numeric

# Import visualization functions
from .plots import (
    plot_bearing_shape,
    plot_key_results,
    plot_load_capacity,
    plot_stiffness,
    plot_pressure_distribution,
    plot_supply_flow_rate,
    plot_chamber_flow_rate,
    plot_ambient_flow_rate,
)

# Import app
from .app.app import app

# Define what should be available when using 'from openairbearing import *'
__all__ = [
    # App
    "app",
    # Bearing types
    "RectangularBearing",
    "CircularBearing",
    "AnnularBearing",
    "InfiniteLinearBearing",
    "JournalBearing",
    # Bearing parameters
    "get_kappa",
    "get_Qsc",
    "get_beta",
    "get_geom",
    "get_area",
    # Result type
    "Result",
    # Solver
    "solve_bearing",
    "get_pressure_numeric",
    # Configuration
    "DEMO_MODE",
    # Visualization
    "plot_bearing_shape",
    "plot_key_results",
    "plot_load_capacity",
    "plot_stiffness",
    "plot_pressure_distribution",
    "plot_supply_flow_rate",
    "plot_chamber_flow_rate",
    "plot_ambient_flow_rate",
]



import numpy as np
import plotly.subplots as sp
import plotly.graph_objects as go
from scipy.interpolate import griddata

# Plot styling
PLOT_FONT = dict(
    family="Arial",
    size=12,
)

# Define solver colors at the top level with other constants
SOLVER_COLORS = {
    "analytic": "blue",
    "numeric": "red",
    "numeric2d": "red",
}

# Common axis properties
AXIS_STYLE = dict(
    title_font=PLOT_FONT,
    tickfont=PLOT_FONT,
    showline=True,
    linecolor="black",
    ticks="inside",
    mirror=True,
)
FIG_LAYOUT = dict(
    font=PLOT_FONT,
    plot_bgcolor="white",
    paper_bgcolor="white",
    legend=dict(orientation="h", yanchor="bottom", y=1.1, xanchor="center", x=0.5),
    showlegend=True,
)


def plot_key_results(bearing, results):
    """Plot key results."""
    figs = []
    b = bearing

    figs.append(plot_load_capacity(b, results))
    figs.append(plot_stiffness(b, results))
    figs.append(plot_pressure_distribution(b, results))
    figs.append(plot_supply_flow_rate(b, results))
    figs.append(plot_ambient_flow_rate(b, results))
    if b.type == "seal":
        figs.append(plot_chamber_flow_rate(b, results))
    return figs


def plot_bearing_shape(bearing):
    """Plot bearing shapes."""
    figs = []
    b = bearing

    figs.append(plot_xy_shape(b))
    figs.append(plot_xz_shape(b))

    return figs


def plot_load_capacity(bearing, results):
    """Plot the load capacity."""
    results = [results] if not isinstance(results, list) else results
    fig = go.Figure()
    b = bearing

    for result in results:
        color = SOLVER_COLORS.get(result.name, "purple")
        idx_k_max = np.argmax(result.k)

        fig.add_trace(
            go.Scatter(
                x=b.ha.flatten() * 1e6,
                y=result.w,
                name=result.name,
                mode="lines+markers",
                marker=dict(
                    color=color,
                    size=[8 if i == idx_k_max else 0 for i in range(b.nx)],
                    symbol="circle",
                ),
                line=dict(color=color),
            )
        )

    fig.update_xaxes(title_text="h (μm)", range=[0, b.ha_max * 1e6], **AXIS_STYLE)
    fig.update_yaxes(title_text="w (N)", range=[0, None], **AXIS_STYLE)
    fig.update_layout(title="Load Capacity", **FIG_LAYOUT)
    return fig


def plot_stiffness(bearing, results):
    """Plot the stiffness."""
    results = [results] if not isinstance(results, list) else results
    fig = go.Figure()
    b = bearing

    for result in results:
        color = SOLVER_COLORS.get(result.name, "purple")
        idx_k_max = np.argmax(result.k)

        fig.add_trace(
            go.Scatter(
                x=b.ha.flatten() * 1e6,
                y=result.k * 1e-6,
                name=result.name,
                mode="lines+markers",
                marker=dict(
                    color=color,
                    size=[8 if i == idx_k_max else 0 for i in range(b.nx)],
                    symbol="circle",
                ),
                line=dict(color=color),
            )
        )

    fig.update_xaxes(title_text="h (μm)", range=[0, b.ha_max * 1e6], **AXIS_STYLE)
    fig.update_yaxes(title_text="k (N/μm)", range=[0, None], **AXIS_STYLE)
    fig.update_layout(title="Static stiffness", **FIG_LAYOUT)
    return fig


def plot_pressure_distribution(bearing, results):
    """Plot the pressure distribution."""
    results = [results] if not isinstance(results, list) else results
    fig = go.Figure()
    b = bearing

    for result in results:
        color = SOLVER_COLORS.get(result.name, "purple")
        idx_k_max = np.argmax(result.k)

        if result.p.ndim == 2:
            n_plots = 3
            # h_plots = np.linspace(b.ha_min**0.5, b.ha_max**0.5, n_plots) ** 2
            in_h = [1, idx_k_max, b.nh - 1]
            h_plots = b.ha[in_h]
            t_locations = np.round(np.linspace(b.nx, 0, n_plots + 2)[1:-1]).astype(int)
            for h_plot, t_loc in zip(h_plots, t_locations):
                in_h = np.abs(b.ha - h_plot).argmin()
                pressures = (result.p[:, in_h] - b.pa) * 1e-6  # Convert to MPa
                fig.add_trace(
                    go.Scatter(
                        x=b.x * 1e3,
                        y=pressures,
                        mode="lines+text",
                        textposition="top center",
                        text=[
                            f"{h_plot*1e6:.2f} μm" if i == t_loc else None
                            for i in range(b.nx)
                        ],
                        textfont=dict(color=color),
                        name=f"{result.name} {h_plot*1e6:.1f} μm",
                        line=dict(color=color),
                        showlegend=False,
                    ),
                )
                fig.update_xaxes(
                    title_text="r (mm)", range=[0, b.xa * 1e3], **AXIS_STYLE
                )
                fig.update_yaxes(title_text="p (MPa)", range=[0, None], **AXIS_STYLE)
        else:
            pressures = (result.p - b.pa) * 1e-6  # Convert to MPa
            if b.case == "journal":
                fig.add_trace(
                    go.Contour(
                        z=pressures[:, :, np.argmax(result.k)],
                        x=b.theta,
                        y=b.y * 1e3,
                        colorscale="Viridis",
                        zmin=0,
                        zmax=np.max(pressures),
                        contours=dict(
                            coloring="heatmap",
                            showlabels=True,
                            labelfont=dict(size=10, color="white"),
                        ),
                        colorbar=dict(title="p (MPa)", thickness=15),
                        name=result.name,
                    )
                )
            else:
                fig.add_trace(
                    go.Contour(
                        z=pressures[:, :, np.argmax(result.k)],
                        x=b.x * 1e3,
                        y=b.y * 1e3,
                        colorscale="Viridis",
                        zmin=0,
                        zmax=np.max(pressures),
                        contours=dict(
                            coloring="heatmap",
                            showlabels=True,
                            labelfont=dict(size=10, color="white"),
                        ),
                        colorbar=dict(title="p (MPa)", thickness=15),
                        name=result.name,
                    )
                )

            fig.update_xaxes(
                title_text="theta (rad)" if b.case == "journal" else "x (mm)",
                range=(
                    [b.theta.min(), b.theta.max()]
                    if b.case == "journal"
                    else [b.x.min() * 1e3, b.x.max() * 1e3]
                ),
                **AXIS_STYLE,
            )
            fig.update_yaxes(
                title_text="y (mm)",
                range=[b.y.min() * 1e3, b.y.max() * 1e3],
                **AXIS_STYLE,
            )
            steps = []
            for i, h in enumerate(bearing.ha * 1e6):
                step = dict(
                    method="update",
                    args=[
                        {"z": [pressures[:, :, i]]},
                        {
                            "title": f"Pressure Distribution (h = {h:.1f})"
                        },  # Update the title
                    ],
                    label=f"{h:.1f} μm",
                )
                steps.append(step)

            sliders = [
                dict(
                    active=idx_k_max,
                    currentvalue={
                        "prefix": "Pressure distribution plot height: ",
                        "font": {"size": 14},
                    },
                    pad={"t": 50},
                    steps=steps,
                )
            ]
            fig.update_layout(sliders=sliders)

        fig.update_layout(title="Pressure distribution", **FIG_LAYOUT)
    return fig


def plot_supply_flow_rate(bearing, results):
    """Plot the supply flow rate."""
    results = [results] if not isinstance(results, list) else results

    fig = go.Figure()
    b = bearing

    for result in results:
        color = SOLVER_COLORS.get(result.name, "purple")
        idx_k_max = np.argmax(result.k)

        fig.add_trace(
            go.Scatter(
                x=b.ha.flatten() * 1e6,
                y=result.qs,
                name=result.name,
                mode="lines+markers",
                marker=dict(
                    color=color,
                    size=[8 if i == idx_k_max else 0 for i in range(b.nx)],
                    symbol="circle",
                ),
                line=dict(color=color),
            )
        )

    fig.update_xaxes(title_text="h (μm)", range=[0, b.ha_max * 1e6], **AXIS_STYLE)
    fig.update_yaxes(
        title_text="q<sub>s</sub> (L/min)", range=[None, None], **AXIS_STYLE
    )
    fig.update_layout(title="Supply flow", **FIG_LAYOUT)
    return fig


def plot_chamber_flow_rate(bearing, results):
    """Plot the chamber flow rate."""
    results = [results] if not isinstance(results, list) else results
    fig = go.Figure()
    b = bearing

    for result in results:
        color = SOLVER_COLORS.get(result.name, "purple")
        idx_k_max = np.argmax(result.k)
        fig.add_trace(
            go.Scatter(
                x=b.ha.flatten() * 1e6,
                y=result.qc,
                name=result.name,
                mode="lines+markers",
                marker=dict(
                    color=color,
                    size=[8 if i == idx_k_max else 0 for i in range(b.nx)],
                    symbol="circle",
                ),
                line=dict(color=color),
            )
        )

    fig.update_xaxes(title_text="h (μm)", range=[0, b.ha_max * 1e6], **AXIS_STYLE)
    fig.update_yaxes(
        title_text="q<sub>c</sub> (L/min)", range=[None, None], **AXIS_STYLE
    )
    fig.update_layout(title="Chanber flow", **FIG_LAYOUT)
    return fig


def plot_ambient_flow_rate(bearing, results):
    """Plot the ambient flow rate."""
    results = [results] if not isinstance(results, list) else results

    fig = go.Figure()
    b = bearing

    for result in results:
        color = SOLVER_COLORS.get(result.name, "purple")
        idx_k_max = np.argmax(result.k)

        fig.add_trace(
            go.Scatter(
                x=b.ha.flatten() * 1e6,
                y=result.qa,
                name=result.name,
                mode="lines+markers",
                marker=dict(
                    color=color,
                    size=[8 if i == idx_k_max else 0 for i in range(b.nx)],
                    symbol="circle",
                ),
                line=dict(color=color),
            )
        )

    fig.update_xaxes(title_text="h (μm)", range=[0, b.ha_max * 1e6], **AXIS_STYLE)
    fig.update_yaxes(title_text="q<sub>a</sub> (L/min)", range=[0, None], **AXIS_STYLE)
    fig.update_layout(title="ambient flow", **FIG_LAYOUT)
    return fig


# def plot_bearing_shape(bearing):
#     """Create bearing shape visualization

#     Args:
#         bearing: Bearing instance with geometry parameters
#     """
#     b = bearing
#     # Create figure
#     # fig = go.Figure()
#     fig = sp.make_subplots(rows=1, cols=2, subplot_titles=("XY Geometry", "XZ Profile"))

#     plot_xz_shape(b, fig)
#     plot_xy_shape(b, fig)

#     for i in range(1, 3):
#         fig.update_xaxes(AXIS_STYLE, row=1, col=i)
#         fig.update_yaxes(AXIS_STYLE, row=1, col=i)

#     fig.update_layout(
#         font=PLOT_FONT,
#         height=300,
#         showlegend=True,
#         plot_bgcolor="white",
#         paper_bgcolor="white",
#         margin=dict(l=20, r=20, t=20, b=20),
#         legend=dict(orientation="h", yanchor="bottom", y=1.1, xanchor="center", x=0.5),
#     )
#     return fig


def plot_xy_shape(bearing):
    fig = go.Figure()
    b = bearing
    # SHAPE XY
    match b.case:
        case "annular" | "circular":

            # outer circle
            theta = np.linspace(0, 2 * np.pi, 100)
            xa = b.xa * np.cos(theta) * 1e3
            ya = b.xa * np.sin(theta) * 1e3
            fig.add_trace(
                go.Scatter(
                    x=xa,
                    y=ya,
                    fill="toself",
                    fillcolor="lightgrey",
                    line=dict(color="black"),
                    name="Shape",
                    showlegend=False,
                ),
            )

            # inner hole
            if b.case == "annular":
                xc = b.xc * np.cos(theta) * 1e3
                yc = b.xc * np.sin(theta) * 1e3
                fig.add_trace(
                    go.Scatter(
                        x=xc,
                        y=yc,
                        fill="toself",
                        fillcolor="white",
                        line=dict(color="black"),
                        name="Shape",
                        showlegend=False,
                    ),
                )

            # symmetry lines
            fig.add_trace(
                go.Scatter(
                    x=[0, 0],
                    y=[-b.xa * 0.2e3, b.xa * 0.2e3],
                    mode="lines",
                    line=dict(color="gray", width=1, dash="dashdot"),
                    showlegend=False,
                ),
            )
            fig.add_trace(
                go.Scatter(
                    x=[-b.xa * 0.2e3, b.xa * 0.2e3],
                    y=[0, 0],
                    mode="lines",
                    line=dict(color="gray", width=1, dash="dashdot"),
                    showlegend=False,
                ),
            )

        case "infinite":
            # right edge
            fig.add_trace(
                go.Scatter(
                    x=np.array([1, 1]) * b.xa * 1e3,
                    y=np.array([0, 1000]),
                    mode="lines",
                    line=dict(color="black"),
                    name="Shape",
                    showlegend=False,
                ),
            )
            # left edge
            fig.add_trace(
                go.Scatter(
                    x=np.array([0, 0]),
                    y=np.array([0, 1000]),
                    fill="tonextx",
                    fillcolor="lightgrey",
                    mode="lines",
                    line=dict(color="black"),
                    name="Shape",
                    showlegend=False,
                ),
            )

            fig.update_yaxes(range=[0, 1000])

            fig.update_xaxes(
                range=np.array([-0.5, 1.5]) * b.xa * 1e3,
            )

        case "rectangular":
            fig.add_trace(
                go.Scatter(
                    x=np.array([-1, -1, 1, 1, -1]) * b.xa * 0.5e3,
                    y=np.array([-1, 1, 1, -1, -1]) * b.ya * 0.5e3,
                    fill="toself",
                    fillcolor="lightgrey",
                    mode="lines",
                    line=dict(color="black"),
                    name="Shape",
                    showlegend=False,
                ),
            )
            fig.update_yaxes(
                range=np.array([-0.6, 0.6]) * b.ya * 1e3,
            )
            fig.update_xaxes(
                range=np.array([-0.6, 0.6]) * b.xa * 1e3,
                scaleanchor="y",
                scaleratio=1,
            )

        case _:
            pass

    fig.update_xaxes(title="x (mm)", **AXIS_STYLE)
    fig.update_yaxes(
        title="y (mm)",
        scaleanchor=False if b.csys == "cartesian" else "x",
        scaleratio=1,
        **AXIS_STYLE,
    )
    fig.update_layout(title="XY profile", **FIG_LAYOUT)
    return fig


def plot_xz_shape(bearing):
    b = bearing
    fig = go.Figure()

    if b.ny > 1:
        if b.case == "circular" or b.case == "annular":
            zr = (b.geom.T).flatten()
            xr = (b.x[None, :] * np.cos(b.y)[:, None]).flatten()
            yr = (b.x[None, :] * np.sin(b.y)[:, None]).flatten()

            print(xr.shape, yr.shape, zr.shape)
            # Define a regular grid over the data
            x = np.linspace(-b.xa, b.xa, 99)
            y = x
            xg, yg = np.meshgrid(x, y)
            # evaluate the z-values at the regular grid through cubic interpolation
            z = griddata((xr, yr), zr, (xg, yg), method="cubic")
            if b.case == "annular":
                dist = np.sqrt(xg**2 + yg**2)
                z[dist < b.xc] = np.nan
            fig.update_xaxes(
                title_text="x (mm)",
                range=np.array([-1.1, 1.1]) * b.xa * 1e3,
                scaleanchor="y",
            )

            fig.update_yaxes(
                title_text="y (mm)",
                range=np.array([-1.1, 1.1]) * b.xa * 1e3,
            )

        elif b.case == "rectangular":
            z = b.geom.T
            x = b.x
            y = b.y
            fig.update_xaxes(
                title_text="x (mm)",
                range=np.array([-0.6, 0.6]) * b.xa * 1e3,
            )
            fig.update_yaxes(
                title_text="y (mm)",
                range=np.array([-0.6, 0.6]) * b.ya * 1e3,
            )
        elif b.case == "journal":
            z = b.geom.T
            x = b.x / b.xa
            y = b.y
            fig.update_xaxes(
                title_text="theta (rad)",
                range=np.array([-1.1, 1.1]) * np.pi,
                **AXIS_STYLE,
            )
            fig.update_yaxes(
                title_text="y (mm)",
                range=np.array([-0.6, 0.6]) * b.ya * 1e3,
                **AXIS_STYLE,
            )

        fig.add_trace(
            go.Contour(
                z=z * 1e6,
                x=x * 1e3,
                y=y * 1e3,
                colorscale="Viridis",
                zmin=0,
                zmax=b.error * 1e6,
                contours=dict(
                    coloring="heatmap",
                    showlabels=True,
                    labelfont=dict(size=10, color="white"),
                ),
                colorbar=dict(title="(μm)", thickness=15),
                name="Profile",
            ),
        )

    else:
        # PROFILE XZ
        match b.case:
            case "circular":
                x = np.concatenate(([-b.x[-1]], -np.flip(b.x), b.x, [b.x[-1]])) * 1e3
                y = np.concatenate(([100], np.flip(b.geom), b.geom, [100])) * 1e6
                t = ["Bearing<br>" if i == b.nx else None for i in range(b.nx * 2)]
            case "annular":
                x = (
                    np.concatenate(
                        ([-b.x[-1]], -np.flip(b.x), [b.x[1]], [b.x[1]], b.x, [b.x[-1]])
                    )
                    * 1e3
                )
                y = (
                    np.concatenate(
                        ([100], np.flip(b.geom), [100], [100], b.geom, [100])
                    )
                    * 1e6
                )
                t = [
                    (
                        "Bearing<br>"
                        if i == i in [b.nx // 2, 2 + b.nx + b.nx // 2]
                        else None
                    )
                    for i in range(b.nx * 2)
                ]
            case "infinite":
                x = np.concatenate(([b.x[1]], b.x, [b.x[-1]])) * 1e3
                y = np.concatenate(([100], b.geom, [100])) * 1e6
                t = ["Bearing<br>" if i == b.nx // 2 else None for i in range(b.nx)]
            case _:
                pass

        fig.add_trace(
            go.Scatter(
                x=x,
                y=y,
                fill="toself",
                fillcolor="lightgrey",
                mode="lines+text",
                textposition="top center",
                text=t,
                textfont=dict(size=14),
                line=dict(color="black"),
                name="Bearing",
                showlegend=True,
            ),
        )

        # Guide surface XZ
        fig.add_trace(
            go.Scatter(
                x=np.array(
                    [
                        (x[-1] - x[-1] * 1.2 if x[1] == 0 else x[1] * 1.2),
                        (x[1] + x[-1]) / 2,
                        x[-1] * 1.2,
                    ]
                ),  # Convert to mm
                y=np.ones(3) * -0.1,  # Convert to um
                mode="lines+text",
                textposition="bottom center",
                text=[None, "<br>Guide surface", None],
                textfont=dict(size=14),
                line=dict(color="gray"),
                name="Shape",
                showlegend=False,
            ),
        )

        # symmetry line
        if b.csys == "polar":
            fig.add_trace(
                go.Scatter(
                    x=[0, 0],
                    y=[-100, 100],
                    mode="lines",
                    line=dict(color="gray", width=1, dash="dashdot"),
                    showlegend=False,
                ),
            )

        fig.update_xaxes(
            title="x (mm)",
            range=[(x[-1] - x[-1] * 1.1 if x[1] == 0 else x[1] * 1.1), x[-1] * 1.1],
            **AXIS_STYLE,
        )
        fig.update_yaxes(
            title="Shape (μm)",
            range=[-0.5 - 0.3e6 * abs(b.error), 1 + np.max(b.geom) * 1e6],
            **AXIS_STYLE,
        )
        fig.update_layout(title="XZ profile", **FIG_LAYOUT)
    return fig


def empty_figure():
    """Create an empty figure with default layout."""
    fig = go.Figure()
    fig.update_layout(
        xaxis=dict(visible=False),
        yaxis=dict(visible=False),
        plot_bgcolor="White",
        margin=dict(l=0, r=0, t=0, b=0),
    )
    return fig



from dash.dependencies import Input, Output
import dash


from openairbearing.plots import plot_bearing_shape, plot_key_results, empty_figure
from openairbearing.bearings import (
    CircularBearing,
    AnnularBearing,
    InfiniteLinearBearing,
    RectangularBearing,
    JournalBearing,
)
from openairbearing.solvers import solve_bearing
from openairbearing.utils import get_kappa, get_Qsc, get_beta


def get_bearing(case):
    """Return default bearing instance based on case."""
    match case:
        case "circular":
            return CircularBearing
        case "annular":
            return AnnularBearing
        case "infinite":
            return InfiniteLinearBearing
        case "rectangular":
            return RectangularBearing
        case "journal":
            return JournalBearing
        case _:
            raise TypeError("no default bearing defined")


def register_callbacks(app):
    """Register all callbacks for the application."""

    # Add callback to handle parameter updates
    @app.callback(
        [
            *[Output(f"shape-plot-{i}", "figure") for i in range(3)],
            *[Output(f"result-plot-{i}", "figure") for i in range(6)],
            Output("kappa-input", "value", allow_duplicate=True),
            Output("Qsc-input", "value", allow_duplicate=True),
        ],
        [
            Input("case-select", "value"),
            Input("solver-select", "value"),
            Input("pa-input", "value"),
            Input("ps-input", "value"),
            Input("pc-input", "value"),
            Input("rho-input", "value"),
            Input("mu-input", "value"),
            Input("hp-input", "value"),
            Input("xa-input", "value"),
            Input("xc-input", "value"),
            Input("nx-input", "value"),
            Input("ya-input", "value"),
            Input("ny-input", "value"),
            Input("ha-min-input", "value"),
            Input("ha-max-input", "value"),
            Input("nh-input", "value"),
            Input("kappa-input", "value"),
            Input("Qsc-input", "value"),
            Input("error-select", "value"),
            Input("error-input", "value"),
            Input("psi-input", "value"),
        ],
        prevent_initial_call="initial_duplicate",
    )
    def update_bearing(
        case,
        solvers,
        pa_mpa,
        ps_mpa,
        pc_mpa,
        rho,
        mu,
        hp_mm,
        xa_mm,
        xc_mm,
        nx,
        ya_mm,
        ny,
        ha_min_um,
        ha_max_um,
        nh,
        kappa,
        Qsc,
        error_type,
        error_um,
        Psi,
    ):
        """Update bearing parameters and recalculate results.

        Args:
            case: Selected bearing case
            solvers: List of selected solvers
            pa_mpa: Ambient pressure in MPa
            ...

        Returns:
            tuple: Updated figures and values
        """
        ctx = dash.callback_context
        if not ctx.triggered:
            raise dash.exceptions.PreventUpdate

        input_id = ctx.triggered[0]["prop_id"].split(".")[0]

        try:
            # Get default bearing for selected case
            bearing_class = get_bearing(case)

            # Create new bearing instance with appropriate class
            b = bearing_class(
                pa=pa_mpa * 1e6,
                ps=ps_mpa * 1e6,
                pc=pc_mpa * 1e6,
                rho=rho,
                mu=mu,
                hp=hp_mm * 1e-3,
                xa=xa_mm * 1e-3,
                xc=xc_mm * 1e-3,
                ya=ya_mm * 1e-3,
                nx=int(nx),
                ny=int(ny),
                ha_min=ha_min_um * 1e-6,
                ha_max=ha_max_um * 1e-6,
                nh=int(nh),
                error_type=error_type,
                error=error_um * 1e-6,
                Psi=Psi,
            )

            # Update kappa or Qsc based on which input changed
            if input_id == "kappa-input" and kappa is not None:
                b.kappa = kappa
                b.Qsc = get_Qsc(b)
                new_kappa = kappa
                new_Qsc = b.Qsc
            elif input_id == "Qsc-input" and Qsc is not None:
                b.Qsc = Qsc
                b.kappa = get_kappa(b)
                new_kappa = b.kappa
                new_Qsc = Qsc
            else:
                # For other inputs, maintain current values
                b.kappa = kappa if kappa is not None else get_kappa(b)
                b.Qsc = Qsc if Qsc is not None else get_Qsc(b)
                new_kappa = b.kappa
                new_Qsc = b.Qsc

            b.beta = get_beta(b)

            # Calculate results for each selected solver
            results = []

            if "analytic" in solvers:
                results.append(solve_bearing(b, soltype="analytic"))
            if "numeric" in solvers:
                results.append(solve_bearing(b, soltype="numeric"))
            if "numeric2d" in solvers:
                results.append(solve_bearing(b, soltype="numeric2d"))

            shape_figures = plot_bearing_shape(b)
            plot_figures = plot_key_results(b, results)

            return (
                *shape_figures + [empty_figure()] * (3 - len(shape_figures)),
                *plot_figures + [empty_figure()] * (6 - len(plot_figures)),
                new_kappa,
                new_Qsc,
            )

        except Exception as e:
            print(f"Error: {e}")
            return (
                *[empty_figure() for _ in range(3)],
                *[empty_figure() for _ in range(6)],
                dash.no_update,
                dash.no_update,
                dash.no_update,
            )

    @app.callback(
        Output("pc-container", "style"),
        Output("xc-container", "style"),
        Output("ya-container", "style"),
        Output("ny-container", "style"),
        Input("case-select", "value"),
    )
    def toggle_containers(case):
        """Show/hide input containers based on selected case."""
        base_style = {
            "grid-template-columns": "200px 100px 20px",
            "gap": "20px",
            "marginTop": "20px",
            "marginBottom": "20px",
            "align-items": "center",
        }
        match case:
            case "annular":
                return (
                    {**base_style, "display": "grid"},
                    {**base_style, "display": "grid"},
                    {**base_style, "display": "none"},
                    {**base_style, "display": "none"},
                )
            case "infinite":
                return (
                    {**base_style, "display": "grid"},
                    {**base_style, "display": "none"},
                    {**base_style, "display": "none"},
                    {**base_style, "display": "none"},
                )
            case "rectangular":
                return (
                    {**base_style, "display": "grid"},
                    {**base_style, "display": "none"},
                    {**base_style, "display": "grid"},
                    {**base_style, "display": "grid"},
                )
            case "journal":
                return (
                    {**base_style, "display": "grid"},
                    {**base_style, "display": "none"},
                    {**base_style, "display": "grid"},
                    {**base_style, "display": "grid"},
                )
            case _:
                return (
                    {**base_style, "display": "none"},
                    {**base_style, "display": "none"},
                    {**base_style, "display": "none"},
                    {**base_style, "display": "none"},
                )

    @app.callback(
        [Output("solver-select", "options"), Output("solver-select", "value")],
        Input("case-select", "value"),
    )
    def update_solver_options(case):
        """Update available solvers based on the selected case."""

        case_solvers = {
            "circular": [
                {
                    "label": "Analytic",
                    "value": "analytic",
                    "disabled": False,
                },
                {
                    "label": "Numeric",
                    "value": "numeric",
                    "disabled": False,
                },
                {
                    "label": "Numeric 2d",
                    "value": "numeric2d",
                    "disabled": True,
                },
            ],
            "annular": [
                {
                    "label": "Analytic",
                    "value": "analytic",
                    "disabled": False,
                },
                {
                    "label": "Numeric",
                    "value": "numeric",
                    "disabled": False,
                },
                {
                    "label": "Numeric 2d",
                    "value": "numeric2d",
                    "disabled": True,
                },
            ],
            "infinite": [
                {
                    "label": "Analytic",
                    "value": "analytic",
                    "disabled": False,
                },
                {
                    "label": "Numeric",
                    "value": "numeric",
                    "disabled": False,
                },
                {
                    "label": "Numeric 2d",
                    "value": "numeric2d",
                    "disabled": True,
                },
            ],
            "rectangular": [
                {
                    "label": "Analytic",
                    "value": "analytic",
                    "disabled": True,
                },
                {
                    "label": "Numeric",
                    "value": "numeric",
                    "disabled": True,
                },
                {
                    "label": "Numeric 2d",
                    "value": "numeric2d",
                    "disabled": False,
                },
            ],
            "journal": [
                {
                    "label": "Analytic",
                    "value": "analytic",
                    "disabled": True,
                },
                {
                    "label": "Numeric",
                    "value": "numeric",
                    "disabled": True,
                },
                {
                    "label": "Numeric 2d",
                    "value": "numeric2d",
                    "disabled": False,
                },
            ],
        }
        # Default solver selections for each case
        case_defaults = {
            "circular": ["analytic"],
            "annular": ["analytic"],
            "infinite": ["analytic"],
            "rectangular": ["numeric2d"],
            "journal": ["numeric2d"],
        }
        # Get the solvers for the selected case
        solvers = case_solvers.get(case, [])
        defaults = case_defaults.get(case, [])

        return solvers, defaults

    @app.callback(
        [
            Output("rho-input", "value"),
            Output("mu-input", "value"),
            Output("hp-input", "value"),
            Output("xa-input", "value"),
            Output("xc-input", "value"),
            Output("ya-input", "value"),
            Output("kappa-input", "value", allow_duplicate=True),
            Output("Qsc-input", "value", allow_duplicate=True),
            Output("pa-input", "value"),
            Output("pc-input", "value"),
            Output("ps-input", "value"),
            Output("ha-min-input", "value"),
            Output("ha-max-input", "value"),
            Output("nx-input", "value"),
            Output("ny-input", "value"),
            Output("nh-input", "value"),
            Output("error-input", "value"),
            Output("psi-input", "value"),
        ],
        [
            Input("reset-all", "n_clicks"),
            Input("rho-reset", "n_clicks"),
            Input("mu-reset", "n_clicks"),
            Input("hp-reset", "n_clicks"),
            Input("xa-reset", "n_clicks"),
            Input("xc-reset", "n_clicks"),
            Input("ya-reset", "n_clicks"),
            Input("kappa-reset", "n_clicks"),
            Input("Qsc-reset", "n_clicks"),
            Input("pa-reset", "n_clicks"),
            Input("pc-reset", "n_clicks"),
            Input("ps-reset", "n_clicks"),
            Input("ha-min-reset", "n_clicks"),
            Input("ha-max-reset", "n_clicks"),
            Input("nx-reset", "n_clicks"),
            Input("ny-reset", "n_clicks"),
            Input("nh-reset", "n_clicks"),
            Input("error-reset", "n_clicks"),
            Input("psi-reset", "n_clicks"),
            Input("case-select", "value"),
        ],
        prevent_initial_call=True,
    )
    def reset_values(reset_all, *args):
        ctx = dash.callback_context
        if not ctx.triggered:
            raise dash.exceptions.PreventUpdate

        button_id = ctx.triggered[0]["prop_id"].split(".")[0]

        # Fetch defaults based on current case selection
        case = args[-1]
        b = get_bearing(case)()

        # Current values dictionary with proper unit conversions
        current_values = {
            "rho": b.rho,
            "mu": b.mu,
            "hp": b.hp * 1e3,  # Convert to mm
            "xa": b.xa * 1e3,  # Convert to mm
            "xc": b.xc * 1e3,  # Convert to mm
            "ya": b.ya * 1e3,  # Convert to mm
            "kappa": b.kappa,
            "Qsc": b.Qsc,
            "pa": b.pa * 1e-6,  # Convert to MPa
            "pc": b.pc * 1e-6,  # Convert to MPa
            "ps": b.ps * 1e-6,  # Convert to MPa
            "ha_min": b.ha_min * 1e6,  # Convert to μm
            "ha_max": b.ha_max * 1e6,  # Convert to μm
            "nx": b.nx,
            "ny": b.ny,
            "nh": b.nh,
            "error": b.error * 1e6,
            "psi": b.Psi,
        }

        # Reset all values when Reset All clicked OR when case changes
        if button_id in ["reset-all", "case-select"]:
            return list(current_values.values())

        # For individual reset buttons
        param = button_id.replace("-reset", "")
        return [
            current_values[p] if p == param else dash.no_update
            for p in [
                "rho",
                "mu",
                "hp",
                "xa",
                "xc",
                "ya",
                "kappa",
                "Qsc",
                "pa",
                "pc",
                "ps",
                "ha_min",
                "ha_max",
                "nx",
                "ny",
                "nh",
                "error",
                "psi",
            ]
        ]



import pytest
import numpy as np
from openairbearing.utils import (
    get_area,
    get_geom,
    get_beta,
    get_kappa,
    get_Qsc,
    get_dA,
    get_load_capacity,
    get_stiffness,
)
from openairbearing.bearings import BaseBearing


def test_get_area():
    """Test the get_area function for different bearing cases."""
    bearing = BaseBearing(case="circular", xa=10)
    assert get_area(bearing) == pytest.approx(np.pi * 10**2)

    bearing = BaseBearing(case="annular", xa=10, xc=5)
    assert get_area(bearing) == pytest.approx(np.pi * (10**2 - 5**2))

    bearing = BaseBearing(case="infinite", xa=10)
    assert get_area(bearing) == pytest.approx(10)

    bearing = BaseBearing(case="rectangular", xa=10, ya=5)
    assert get_area(bearing) == pytest.approx(10 * 5)

    with pytest.raises(ValueError, match="Unknown case: unknown"):
        bearing = BaseBearing(case="unknown")
        get_area(bearing)


def test_get_geom():
    """Test the get_geom function for different error types."""
    nx = 10
    ny = 5

    bearing = BaseBearing(
        case="circular", csys="polar", nx=nx, xa=10, ya=5, error_type="none"
    )
    geom = get_geom(bearing)
    assert np.all(geom == 0)

    bearing.error_type = "linear"
    geom = get_geom(bearing)
    assert geom.shape == (nx,)

    bearing.error_type = "quadratic"
    geom = get_geom(bearing)
    assert geom.shape == (nx,)

    with pytest.raises(ValueError, match="Unknown error type: invalid"):
        bearing.error_type = "invalid"
        get_geom(bearing)

    bearing = BaseBearing(
        case="rectangular", nx=nx, ny=ny, xa=10, ya=5, error_type="none"
    )
    geom = get_geom(bearing)
    assert np.all(geom == 0)

    bearing.error_type = "linear"
    geom = get_geom(bearing)
    assert geom.shape == (nx, ny)

    bearing.error_type = "quadratic"
    geom = get_geom(bearing)
    assert geom.shape == (nx, ny)

    with pytest.raises(ValueError, match="Unknown error type: invalid"):
        bearing.error_type = "invalid"
        get_geom(bearing)


def test_get_beta():
    """Test the get_beta function."""
    ha = np.linspace(1e-6, 20e-6, 10)
    hp = 4e-3
    kappa = 1e-15
    xa = 15
    bearing = BaseBearing(case="circular", xa=xa, hp=hp)
    bearing.kappa = kappa
    bearing.ha = ha
    beta = get_beta(bearing)
    assert beta.shape == ha.shape
    assert beta == pytest.approx(6 * kappa * xa**2 / (hp * ha**3))


def test_get_kappa():
    """Test the get_kappa function."""
    bearing = BaseBearing(
        case="circular", Qsc=3, mu=1.85e-5, hp=1e-3, pa=101325, psc=0.6e6 + 101325
    )
    bearing.A = get_area(bearing)
    kappa = get_kappa(bearing)
    assert kappa > 0


def test_get_Qsc():
    """Test the get_Qsc function."""
    bearing = BaseBearing(case="circular", psc=0.6e6)
    bearing.kappa = 1e-15
    Qsc = get_Qsc(bearing)
    assert Qsc > 0


def test_get_dA():
    """Test the get_dA function."""
    bearing = BaseBearing(case="circular", csys="polar", nx=10, ny=1, xa=10)
    dA = get_dA(bearing)
    bearing = BaseBearing(case="infinite", csys="cartesian", nx=10, ny=1, xa=10)
    dA = get_dA(bearing)
    assert dA.shape == (10,)


# def test_get_load_capacity():
#     """Test the get_load_capacity function."""
#     bearing = BaseBearing(case="circular", type="bearing", csys="polar", nx=10, ny=1, xa=10, pa=101325)
#     p = np.ones(10) * 101325
#     w = get_load_capacity(bearing, p)
#     assert w.shape == p.shape ,  f"w {w.shape}, p {p.shape}"
#     # assert np.allclose(w, np.sum (p * bearing.dA, axis=0))


def test_get_stiffness():
    """Test the get_stiffness function."""
    nh = 40
    bearing = BaseBearing(case="circular", ha_min=1e-6, ha_max=20e-6, nh=40)
    w = 2 ** np.linspace(0, 100, nh)
    k = get_stiffness(bearing, w)
    assert np.allclose(k, -np.gradient(w, bearing.ha.flatten()))



import dash

from openairbearing.app.layouts import create_layout
from openairbearing.app.callbacks import register_callbacks
from openairbearing.bearings import CircularBearing
from openairbearing.solvers import solve_bearing

# Define input fields with default values
default_bearing = CircularBearing()

# Initialize bearing and calculate performance
bearing = default_bearing
result = solve_bearing(bearing, "analytic")

app = dash.Dash(
    __name__,
    meta_tags=[
        {"name": "viewport", "content": "width=device-width, initial-scale=1.0"}
    ],
    title="OpenAirBearing",
    update_title="Loading...",
    assets_folder="assets",
)

# Update app.layout before app.run_server(debug=True):
app.layout = create_layout(default_bearing, bearing, result)

# Register callbacks
register_callbacks(app)

app.index_string = """
<!DOCTYPE html>
<html>
    <head>
        {%metas%}
        <title>{%title%}</title>
        <link rel="shortcut icon" href="/assets/favicon.ico">
        {%css%}
    </head>
    <body>
        {%app_entry%}
        <footer>
            <p> Open Air Bearing is available as open-source software under MIT license: <a href="https://github.com/Aalto-Arotor/openAirBearing">github.com/Aalto-Arotor/openAirBearing</a> - Contact: <a href="mailto:mikael.miettinen@iki.fi">mikael.miettinen@iki.fi</a></p>
            {%config%}
            {%scripts%}
            {%renderer%}
        </footer>
    </body>
</html>
"""


def main():
    app.run_server(debug=True)


if __name__ == "__main__":
    main()



import pytest
import numpy as np

from openairbearing.bearings import *
from openairbearing.solvers import (
    solve_bearing,
    get_pressure_analytic_circular,
    get_pressure_analytic_annular,
    get_pressure_analytic_infinite,
    get_pressure_numeric,
    get_pressure_2d_numeric,
)


def test_get_pressure_analytic_circular():
    """Test the analytic pressure distribution for a circular bearing."""
    bearing = CircularBearing()
    p = get_pressure_analytic_circular(bearing)
    assert p.shape == (bearing.nx, bearing.nh)
    e = 1e-6
    print(np.min(p - bearing.ps), np.max(p - bearing.ps))
    assert np.all(p - bearing.pa > -e)
    assert np.all(p - bearing.ps < e)


def test_get_pressure_analytic_annular():
    """Test the analytic pressure distribution for a annular bearing."""
    bearing = AnnularBearing()
    p = get_pressure_analytic_annular(bearing)
    assert p.shape == (bearing.nx, bearing.nh)
    assert p[0, :] == pytest.approx(bearing.pc)
    assert p[-1, :] == pytest.approx(bearing.pa)
    e = 1e-6
    assert np.all(p - bearing.pa > -e)
    assert np.all(p - bearing.ps < e)
    # assert np.all(p <= bearing.ps)


def test_get_pressure_analytic_infinite():
    """Test the analytic pressure distribution for a infinite linear bearing."""
    bearing = InfiniteLinearBearing()
    p = get_pressure_analytic_infinite(bearing)
    assert p.shape == (bearing.nx, bearing.nh)
    e = 1e-6
    assert np.all(p - bearing.pa > -e)
    assert np.all(p - bearing.ps < e)


def test_get_pressure_numeric():
    """Test the numeric pressure distribution for a circular bearing."""
    bearing = CircularBearing()
    p = get_pressure_numeric(bearing)
    assert p.shape == (bearing.nx, bearing.nh)
    e = 1e-6
    assert np.all(p - bearing.pa > -e)
    assert np.all(p - bearing.ps < e)


# def test_get_pressure_2d_numeric():
#     """Test the numeric pressure distribution for a rectangular bearing."""
#     bearing = RectangularBearing()
#     p = get_pressure_2d_numeric(bearing)
#     assert p.shape == (bearing.ny, bearing.nx, bearing.nh)
#     e = 1e-6
#     assert np.all(p - bearing.pa > -e)
#     assert np.all(p - bearing.ps < e)


def test_solve_bearing():
    """Test the solve_bearing function."""
    par = {
        "nh": 200,
        "ha_min": 1e-6,
        "ha_max": 100e-6,
        "nx": 100,
        "error_type": "none",
        "error": 0,
        "Psi": 0,
    }
    # test that analytical solutions match the numerical solutions with tolerance e
    for bearing in [
        CircularBearing(**par),
        AnnularBearing(**par),
        InfiniteLinearBearing(**par),
    ]:
        results = [
            solve_bearing(bearing, "analytic"),
            solve_bearing(bearing, "numeric"),
        ]
        e = 0.05
        assert np.allclose(results[0].p, results[1].p, rtol=e)
        assert np.allclose(results[0].w, results[1].w, rtol=e)
        assert np.allclose(results[0].k, results[1].k, rtol=e)
        assert np.allclose(results[0].qs, results[1].qs, rtol=e)



from dash import html, dcc

from openairbearing.plots import plot_bearing_shape, plot_key_results, empty_figure
from openairbearing.config import DEMO_MODE


STYLES = {
    "input": {
        "width": "100px",
        "borderRadius": "4px",
        "border": "1px solid #ddd",
        "padding": "4px",
    },
    "input_container": {
        "display": "grid",
        "grid-template-columns": "200px 100px 20px",
        "marginBottom": "20px",
        "gap": "20px",
        "align-items": "center",
    },
    "toggle_container": {
        "display": "none",
        "grid-template-columns": "200px 100px 20px",
        "marginTop": "20px",
        "gap": "20px",
        "align-items": "center",
    },
    "reset_button": {
        "padding": "2px 6px",
        "fontSize": "14px",
        "backgroundColor": "#f8f9fa",
        "border": "1px solid #ddd",
        "borderRadius": "4px",
        "cursor": "pointer",
        "height": "25px",
        "width": "30px",
    },
    "header_container": {
        "display": "flex",
        "justifyContent": "space-between",
        "alignItems": "center",
        "marginBottom": "20px",
    },
    "reset_all_button": {
        "padding": "5px 10px",
        "fontSize": "14px",
        "backgroundColor": "#f8f9fa",
        "border": "1px solid #ddd",
        "borderRadius": "4px",
        "cursor": "pointer",
        "height": "30px",
    },
    "input_column": {
        "width": "370px",
        "minWidth": "370px",  # Prevent inputs from getting too narrow
        "display": "inline-block",
        "vertical-align": "top",
        "padding": "20px",
        "border": "1px solid black",
        "borderRadius": "8px",
        "flex": "0 1 auto",  # Don't grow, allow shrink, auto basis
    },
    "plot_box": {
        "width": "calc(100% - 40px)",  # Dynamic width based on container
        "display": "inline-block",
        "vertical-align": "top",
        "justifyContent": "space-between",
        "padding": "20px",
        "border": "1px solid black",
        "borderRadius": "8px",
        "flex": "1 1 auto",  # Allow grow and shrink
    },
    "plot_column": {
        "width": "calc(100% - 550px)",  # Dynamic width based on container
        "minWidth": "600px",  # Minimum width for plots
        "display": "inline-block",
        "vertical-align": "top",
        "padding": "0px",
        "flex": "1 1 auto",  # Allow grow and shrink
    },
    "container": {
        "display": "flex",
        "alignItems": "flex-start",
        "justifyContent": "space-between",
        "width": "100%",
        "flexWrap": "wrap",  # Allow wrapping on smaller screens
        "gap": "20px",  # Space between columns
    },
}


def create_layout(default_bearing, bearing, results):
    """Create the main app layout.

    Args:
        bearing: Bearing instance
        results: List of calculation results

    Returns:
        html.Div: Main application layout
    """
    return html.Div(
        [
            html.Div(
                [
                    html.Img(
                        src="/assets/favicon.ico",
                        style={
                            "height": "40px",
                            "margin": "10px 5px",
                            "verticalAlign": "middle",
                        },
                    ),
                    html.H1(
                        "Open Air Bearing",
                        style={
                            "textAlign": "center",
                            "display": "inline-block",
                            "margin": "10px 0",
                        },
                    ),
                ],
                style={
                    "display": "flex",
                    "alignItems": "center",
                    "justifyContent": "center",
                },
            ),
            html.Div(
                [
                    create_input_layout(default_bearing),
                    create_results_layout(bearing, results),
                ],
                style=STYLES["container"],
            ),
        ]
    )


def create_input_layout(default_bearing):
    return html.Div(
        [
            html.Div(
                [
                    html.H3("Setup", style={"margin": "0"}),
                    html.Button(
                        "Reset All",
                        id="reset-all",
                        title="Reset all values to default",
                        style=STYLES["reset_all_button"],
                    ),
                ],
                style=STYLES["header_container"],
            ),
            # html.H4("Setup"),
            html.Div(
                [
                    html.Label("Simulated Case"),
                    dcc.Dropdown(
                        id="case-select",
                        options=[
                            {"label": "Circular thrust", "value": "circular"},
                            {"label": "Annular thrust", "value": "annular"},
                            {"label": "Infinitely long", "value": "infinite"},
                            {"label": "Rectangular", "value": "rectangular"},
                            {"label": "Journal", "value": "journal", "disabled": True},
                        ],
                        value="circular",
                        style={"width": "150px"},
                    ),
                    html.Label(""),
                    html.Label("Solution selection"),
                    dcc.Checklist(
                        id="solver-select",
                        options=[
                            {
                                "label": "Analytic",
                                "value": "analytic",
                                "disabled": False,
                            },
                            {
                                "label": "Numeric",
                                "value": "numeric",
                                "disabled": False,
                            },
                            {
                                "label": "Numeric 2d",
                                "value": "numeric2d",
                                "disabled": False,
                            },
                        ],
                        value=["analytic"],
                        style={"color": "black", "width": "150px"},
                    ),
                ],
                style=STYLES["input_container"],
            ),
            # Geometry inputs
            html.H4("Bearing parameters"),
            html.Div(
                [
                    html.Label("Porous Layer Thickness (mm)"),
                    dcc.Input(
                        id="hp-input",
                        type="number",
                        min=0.01,
                        value=default_bearing.hp * 1e3,  # Convert m to mm
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="hp-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                    html.Label("Outer radius / length (mm)"),
                    dcc.Input(
                        id="xa-input",
                        type="number",
                        min=0.01,
                        value=default_bearing.xa * 1e3,  # Convert m to mm
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="xa-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                style=STYLES["input_container"],
            ),
            html.Div(
                [
                    html.Label("Inner radius (mm)"),
                    dcc.Input(
                        id="xc-input",
                        type="number",
                        min=0.01,
                        value=default_bearing.xc * 1e3,  # Convert m to mm
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="xc-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                id="xc-container",
                style=STYLES["toggle_container"],
            ),
            html.Div(
                [
                    html.Label("Length (mm)"),
                    dcc.Input(
                        id="ya-input",
                        type="number",
                        min=0.01,
                        value=default_bearing.ya * 1e3,  # Convert m to mm
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="ya-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                id="ya-container",
                style=STYLES["toggle_container"],
            ),
            html.Div(
                [
                    html.Label("Permeability (m^2)"),
                    dcc.Input(
                        id="kappa-input",
                        type="number",
                        value=default_bearing.kappa,
                        min=0,
                        step=1e-16,
                        inputMode="numeric",
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="kappa-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                    html.Label("Free flow (l/min)"),
                    dcc.Input(
                        id="Qsc-input",
                        type="number",
                        value=default_bearing.Qsc,
                        min=0.1,
                        step=0.1,
                        inputMode="numeric",
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="Qsc-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                style=STYLES["input_container"],
            ),
            # Geometry inputs
            html.H4("Numerical model specific:"),
            html.Div(
                [
                    html.Label("Geometrical error type"),
                    dcc.Dropdown(
                        id="error-select",
                        options=[
                            {
                                "label": "Linear",
                                "value": "linear",
                            },
                            {
                                "label": "Quadratic",
                                "value": "quadratic",
                            },
                            {
                                "label": "Tilt x",
                                "value": "tiltx",
                                "disabled": True,
                            },
                            {
                                "label": "Tilt y",
                                "value": "tilty",
                                "disabled": True,
                            },
                        ],
                        value="linear",
                        style={"width": "150px"},
                    ),
                    html.Label(""),
                    html.Label("Geometry error (μm)"),
                    dcc.Input(
                        id="error-input",
                        type="number",
                        step=0.5,
                        value=default_bearing.error * 1e6,
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="error-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                    html.Label("Slip coefficeint Φ"),
                    dcc.Input(
                        id="psi-input",
                        type="number",
                        min=0,
                        step=0.01,
                        value=default_bearing.Psi * 1e6,
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="psi-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                style=STYLES["input_container"],
            ),
            html.H4("Load parameters"),
            html.Div(
                [
                    html.Label("Ambient Pressure (MPa)"),
                    dcc.Input(
                        id="pa-input",
                        type="number",
                        value=default_bearing.pa * 1e-6,  # Convert Pa to MPa
                        min=0,
                        step=0.1,
                        inputMode="numeric",
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="pa-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                    html.Label("Supply Pressure (MPa)"),
                    dcc.Input(
                        id="ps-input",
                        type="number",
                        value=default_bearing.ps * 1e-6,  # Convert Pa to MPa
                        min=0.1,
                        step=0.1,
                        inputMode="numeric",
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="ps-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                style=STYLES["input_container"],
            ),
            html.Div(
                [
                    html.Label("Chamber Pressure (MPa)"),
                    dcc.Input(
                        id="pc-input",
                        type="number",
                        value=default_bearing.pc * 1e-6,  # Convert Pa to MPa
                        min=0.1,
                        step=0.1,
                        inputMode="numeric",
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="pc-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                id="pc-container",
                style=STYLES["toggle_container"],
            ),
            html.H4("Fluid properties"),
            html.Div(
                [
                    html.Label("Air Density (kg/m³)"),
                    dcc.Input(
                        id="rho-input",
                        type="number",
                        value=default_bearing.rho,
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="rho-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                    html.Label("Dynamic Viscosity (Pa·s)"),
                    dcc.Input(
                        id="mu-input",
                        type="number",
                        value=default_bearing.mu,
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="mu-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                style=STYLES["input_container"],
            ),
            html.H4("Model parameters"),
            html.Div(
                [
                    html.Label("Minimum Height (μm)"),
                    dcc.Input(
                        id="ha-min-input",
                        type="number",
                        value=default_bearing.ha_min * 1e6,  # Convert m to μm
                        min=0,
                        step=0.5,
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="ha-min-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                    html.Label("Maximum Height (μm)"),
                    dcc.Input(
                        id="ha-max-input",
                        type="number",
                        value=default_bearing.ha_max * 1e6,  # Convert m to μm
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="ha-max-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                    html.Label("Number of height points"),
                    dcc.Input(
                        id="nh-input",
                        type="number",
                        min=3,
                        max=100 if DEMO_MODE else None,
                        step=1,
                        value=default_bearing.nh,
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="nh-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                    html.Label("Number of x direction points"),
                    dcc.Input(
                        id="nx-input",
                        type="number",
                        value=default_bearing.nx,
                        min=3,
                        max=100 if DEMO_MODE else None,
                        step=1,
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="nx-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                style=STYLES["input_container"],
            ),
            html.Div(
                [
                    html.Label("Number of y direction points"),
                    dcc.Input(
                        id="ny-input",
                        type="number",
                        value=default_bearing.ny,
                        min=3,
                        max=100 if DEMO_MODE else None,
                        step=1,
                        style=STYLES["input"],
                    ),
                    html.Button(
                        "↺",
                        id="ny-reset",
                        title="Reset to default",
                        style=STYLES["reset_button"],
                    ),
                ],
                id="ny-container",
                style=STYLES["toggle_container"],
            ),
        ],
        style=STYLES["input_column"],
    )


def create_results_layout(bearing, results):
    """Create the results section layout.

    Args:
        bearing: Bearing instance
        results: List of calculation results

    Returns:
        html.Div: Results layout with plots arranged in a grid.
    """
    # Get the list of plots from plot_key_results
    shape_figures = plot_bearing_shape(bearing)
    plot_figures = plot_key_results(bearing, results)

    # Pad the list to ensure it has a multiple of 3 elements
    while len(shape_figures) % 3 != 0:
        shape_figures.append(empty_figure())
    while len(plot_figures) % 3 != 0:
        plot_figures.append(empty_figure())
    # Create rows of plots (3 plots per row)
    shape_rows = []
    for i in range(0, len(shape_figures), 3):
        row = html.Div(
            [
                html.Div(
                    dcc.Graph(
                        id=f"shape-plot-{j}",
                        figure=shape_figures[j],
                        config={"displayModeBar": False},
                        style={"height": "400px"},
                    ),
                    style={
                        "width": "calc(33% - 20px)",
                        "margin": "10px",
                        "padding": "10px",
                    },
                )
                for j in range(
                    i, min(i + 3, len(shape_figures))
                )  # Add up to 3 plots per row
            ],
            style={
                "display": "flex",
                "justifyContent": "space-between",
            },  # Flexbox for row layout
        )
        shape_rows.append(row)
    result_rows = []
    for i in range(0, len(plot_figures), 3):
        row = html.Div(
            [
                html.Div(
                    dcc.Graph(
                        id=f"result-plot-{j}",
                        figure=plot_figures[j],
                        config={"displayModeBar": False},
                        style={"height": "400px"},
                    ),
                    style={
                        "width": "calc(33% - 20px)",
                        "margin": "10px",
                        "padding": "10px",
                    },
                )
                for j in range(
                    i, min(i + 3, len(plot_figures))
                )  # Add up to 3 plots per row
            ],
            style={
                "display": "flex",
                "justifyContent": "space-between",
            },  # Flexbox for row layout
        )
        result_rows.append(row)

    # Combine all rows into a single column
    return html.Div(
        [
            html.Div(
                [
                    html.H3(
                        "Bearing shape",
                        style={"margin": "0 0 20px 0", "textAlign": "left"},
                    ),
                    *shape_rows,
                ],
                style=STYLES["plot_box"],
            ),
            html.Div(style={"height": "20px"}),
            html.Div(
                [
                    html.H3(
                        "Results", style={"margin": "0 0 20px 0", "textAlign": "left"}
                    ),
                    *result_rows,
                ],
                style=STYLES["plot_box"],
            ),
        ],
        style=STYLES["plot_column"],  # Style for the overall results section
    )


# def create_results_layout(bearing, results):
#     """Create the results section layout.

#     Args:
#         bearing: Bearing instance
#         results: List of calculation results
#     """
#     return html.Div(
#         [
#             html.Div(
#                 [
#                     html.H3("Bearing shape", style={"margin": "0"}),
#                     dcc.Graph(
#                         id="bearing-shape",
#                         figure=plot_bearing_shape(bearing),
#                         config={"displayModeBar": False},
#                     ),
#                 ],
#                 style=STYLES["plot_box"],
#             ),
#             # Spacer
#             html.Div(style={"height": "20px"}),
#             html.Div(
#                 [
#                     html.H3("Results", style={"margin": "0"}),
#                     dcc.Graph(
#                         id="bearing-plots",
#                         figure=plot_key_results(bearing, results),
#                         config={"displayModeBar": False},
#                     ),
#                 ],
#                 style=STYLES["plot_box"],
#             ),
#         ],
#         style=STYLES["plot_column"],
#     )



from dataclasses import dataclass

import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from scipy.special import i0, k0

from openairbearing.utils import (
    Result,
    get_load_capacity,
    get_stiffness,
    get_volumetric_flow,
)


def solve_bearing(bearing, soltype: bool) -> Result:
    match soltype:
        case "analytic":
            name = "analytic"
            match bearing.case:
                case "circular":
                    p = get_pressure_analytic_circular(bearing)
                case "annular":
                    p = get_pressure_analytic_annular(bearing)
                case "infinite":
                    p = get_pressure_analytic_infinite(bearing)
                case _:
                    return Result(
                        name="none",
                        p=np.array([]),
                        w=np.array([]),
                        k=np.array([]),
                        qs=np.array([]),
                        qa=np.array([]),
                        qc=np.array([]),
                    )
        case "numeric":
            name = "numeric"
            p = get_pressure_numeric(bearing)
        case "numeric2d":
            name = "numeric2d"
            p = get_pressure_2d_numeric(bearing)
        case _:
            raise ValueError("Invalid solution type")

    w = get_load_capacity(bearing=bearing, p=p)
    k = get_stiffness(bearing=bearing, w=w)
    qs, qa, qc = get_volumetric_flow(bearing=bearing, p=p, soltype=soltype)

    return Result(name=name, p=p, w=w, k=k, qs=qs, qa=qa, qc=qc)


def get_pressure_analytic_infinite(bearing):
    """
    Calculates the solution for the pressure distribution in infinitely long bearings and seals.
    """

    b = bearing

    f = (2 * b.beta) ** 0.5
    slip = (1 + b.Psi) ** 0.5

    # nondimensionals
    Pa = 1
    Ra = 1
    R = b.x / b.xa
    Ps = b.ps / b.pa
    Pc = b.pc / b.pa

    exp_f = np.exp((f * Ra) / slip)

    numer1 = -(Pc**2) + Ps**2 + exp_f * (Pa**2 - Ps**2)
    numer2 = exp_f * (-(Pa**2) + Ps**2 + exp_f * (Pc**2 - Ps**2))

    denom = -1 + np.exp((2 * f * Ra) / slip)

    C1 = numer1 / denom
    C2 = numer2 / denom

    p = (
        b.pa
        * (
            Ps**2
            + C1 * np.exp(np.outer(R, f) / slip)
            + C2 * np.exp(-np.outer(R, f) / slip)
        )
        ** 0.5
    )
    return p


def get_pressure_analytic_annular(bearing):
    """
    Calculates the Bessel function solution for the pressure distribution in annular bearings and seals.
    """

    b = bearing

    f = (2 * b.beta) ** 0.5

    # nondimensionals
    Pa = 1
    Ra = 1
    R = b.x / b.xa
    Ps = b.ps / b.pa
    Pc = b.pc / b.pa
    Rc = b.xc / b.xa

    numer1 = (Pa**2 - Ps**2) * k0(f * Rc) + (Ps**2 - Pc**2) * k0(f * Ra)
    numer2 = (Pa**2 - Ps**2) * i0(f * Rc) + (Ps**2 - Pc**2) * i0(f * Ra)

    denom = i0(f * Rc) * k0(f * Ra) - i0(f * Ra) * k0(f * Rc)

    C1 = numer1 / denom
    C2 = numer2 / denom

    p = b.pa * (Ps**2 - C1 * i0(np.outer(R, f)) + C2 * k0(np.outer(R, f))) ** 0.5
    return p


def get_pressure_analytic_circular(bearing):
    """
    Calculates the Bessel function solution for the pressure distribution in circluar trust bearings.
    """
    b = bearing
    p = (
        b.ps
        * (
            1
            - (1 - b.pa**2 / b.ps**2)
            * i0(np.outer(b.x / b.xa, (2 * b.beta) ** 0.5))
            / i0((2 * b.beta) ** 0.5)
        )
        ** 0.5
    )
    return p


def get_pressure_numeric(bearing, ha=None, ps=None):
    b = bearing

    # uniform kappa
    kappa = b.kappa * np.ones_like(b.x)

    # Partially blocked restrictors, set blocked region to 0 permeability
    if b.blocked:
        kappa[b.block_in] = 0

    # porous feeding terms
    porous_source = -kappa / (2 * b.hp * b.mu)

    if ha is None:
        ha = b.ha
    else:
        ha = np.atleast_1d(ha)
    if ps is None:
        ps = b.ps

    p = np.zeros((len(b.x), len(ha)))

    for i in range(len(ha)):
        h = ha[i] + b.geom

        if b.csys == "polar":
            epsilon = (1 + b.Psi) * b.x * h**3 / (24 * b.mu)
            coefficient = sp.diags(1 / b.x, 0)
        elif b.csys == "cartesian":
            epsilon = (1 + b.Psi) * h**3 / (24 * b.mu)
            coefficient = 1

        diff_mat = build_diff_matrix(coefficient, epsilon, b.dx)
        A = sp.lil_matrix(diff_mat + sp.diags(porous_source, 0))

        f = ps**2 * porous_source

        # Boundary conditions
        if b.type == "bearing":
            # Neumann at r=0
            A[0, 1] = -A[0, 0]
            f[0] = 0
        elif b.type == "seal":
            # Dirilecht at r=rc
            A[0, 0] = 1
            A[0, 1] = 0
            f[0] = b.pc**2

        # Dirichlet at r=ra
        A[-1, -2] = 0
        A[-1, -1] = 1
        f[-1] = b.pa**2

        A = A.tocsr()
        p[:, i] = spla.spsolve(A, f) ** 0.5

    return p


def build_diff_matrix(
    coef: np.ndarray, eps: np.ndarray, dr: np.ndarray
) -> sp.csr_matrix:
    """Construct finite-difference matrix for coefficient @ D_r(epsilon * D_r(f(r)))

    Builds a sparse matrix representing the discretized differential operator
    using second-order central differences with variable coefficients.
    """

    N = len(dr)

    # Compute epsilon at half-points
    eps_half = (eps[:-1] + eps[1:]) / 2
    # Finite difference second derivative matrix with variable coefficient
    diag_main = np.zeros(N)
    diag_upper = np.zeros(N - 1)
    diag_lower = np.zeros(N - 1)

    # interior points with 3 point stencil
    diag_main[1:-1] = -(eps_half[1:] + eps_half[:-1]) / dr[1:-1] ** 2
    diag_upper[1:] = eps_half[1:] / dr[1:-1] ** 2
    diag_lower[:-1] = eps_half[:-1] / dr[1:-1] ** 2

    # Assemble sparse matrix
    L_mat = sp.diags([diag_lower, diag_main, diag_upper], [-1, 0, 1], format="csr")
    # Handle also scalar coefficients
    if isinstance(coef, (float, int)):
        return coef * L_mat
    else:
        return coef @ L_mat


def get_pressure_2d_numeric(bearing):
    """
    Solve the 2D pressure distribution for the bearing using finite differences.

    Args:
        bearing: Bearing object containing geometry and properties.

    Returns:
        np.ndarray: 2D pressure distribution array.
    """
    b = bearing
    N = b.nx
    M = b.ny

    p = np.zeros((M, N, len(b.ha)))

    # Uniform kappa
    kappa = b.kappa * np.ones((N, M))

    # Partially blocked restrictors, set blocked region to 0 permeability
    if b.blocked:
        kappa[b.block_in] = 0

    # Porous feeding terms
    porous_source = -kappa / (2 * b.hp * b.mu)

    for i in range(b.nh):
        if b.case == "journal":
            h = (b.clearance + b.geom[:, :, None])[:, :, i]
        else:
            h = b.ha[i] + b.geom

        if b.csys == "polar":
            epsilon_r = b.x[:, None] * (1 + b.Psi) * h**3 / (24 * b.mu)
            epsilon_theta = (1 + b.Psi) * h**3 / (24 * b.mu)
            epsilon = (epsilon_r, epsilon_theta)
        elif b.csys == "cartesian":
            epsilon = (1 + b.Psi) * h**3 / (24 * b.mu)

        # boundary conditions
        if b.case == "rectangular":
            bc = {
                "west": "Dirichlet",
                "east": "Dirichlet",
                "north": "Dirichlet",
                "south": "Dirichlet",
            }
            bc_vals = {
                "west": b.pa,
                "east": b.pa,
                "north": b.pa,
                "south": b.pa,
            }
            factors = [1, 1]
        elif b.case == "circular":
            bc = {
                "west": "Neumann",
                "east": "Dirichlet",
                "north": "Periodic",
                "south": "Periodic",
            }
            bc_vals = {
                "west": b.pa,
                "east": b.pa,
                "north": b.pa,
                "south": b.pa,
            }
            factors[b.x, b.x**2]
        elif b.case == "annular":
            bc = {
                "west": "Dirichlet",
                "east": "Dirichlet",
                "north": "Periodic",
                "south": "Periodic",
            }
            bc_vals = {
                "west": b.pa,
                "east": b.pc,
            }
            factors[b.x, b.x**2]
        elif b.case == "journal":
            bc = {
                "west": "Dirichlet",
                "east": "Dirichlet",
                "north": "Periodic",
                "south": "Periodic",
            }
            bc_vals = {
                "west": b.pa,
                "east": b.pc,
            }
            factors = [b.xa**2, 1]

        p[:, :, i] = fdm_2d(
            epsilon=epsilon,
            porous_source=porous_source,
            ps=b.ps,
            dx=b.dx,
            dy=b.dy,
            bc=bc,
            bc_vals=bc_vals,
            N=N,
            M=M,
            factors=factors,
        )
    return p


def fdm_2d(
    epsilon: np.ndarray,
    porous_source: np.ndarray,
    ps: float,
    dx: np.ndarray,
    dy: np.ndarray,
    bc: dict,
    bc_vals: dict,
    N: int,
    M: int,
    factors: list = [1, 1],
) -> np.ndarray:
    """
    Solve air gap pressure in 2D using a finite difference scheme.

    Args:
        epsilon (np.ndarray): Coefficient matrix (N, M).
        porous_source (float): Source term through porous restrictor.
        ps (float): Supply pressure
        dx (float): Grid spacing in the x-direction.
        dy (float): Grid spacing in the y-direction.
        bc (dict): Boundary conditions for "west", "east", "north", "south".
                   Each can be "Dirichlet", "Neumann", or "Periodic".
        dirichlet_values (dict): Dirichlet values for "west", "east", "north", "south".
        N (int): Number of grid points in the x-direction.
        M (int): Number of grid points in the y-direction.

    Returns:
        np.ndarray: Solution pressure matrix p (N, M).
    """
    # Number of unknowns
    num_points = N * M
    fx = factors[0]
    fy = factors[1]

    # Create sparse matrix and right-hand side
    A = sp.lil_matrix((num_points, num_points))

    # edge source = 0 except for periodic boundaries
    if bc["west"] != "Periodic":
        porous_source[:, 0] = 0
    if bc["east"] != "Periodic":
        porous_source[:, -1] = 0
    if bc["north"] != "Periodic":
        porous_source[-1, :] = 0
    if bc["south"] != "Periodic":
        porous_source[0, :] = 0

    b = porous_source.flatten() * ps**2

    # Helper function to convert 2D indices to 1D
    def idx(i, j):
        return i * M + j

    # Compute epsilon at half-points
    eps_w = np.zeros_like(epsilon)
    eps_e = np.zeros_like(epsilon)
    eps_n = np.zeros_like(epsilon)
    eps_s = np.zeros_like(epsilon)

    # West/east half-points
    eps_w[:, :-1] = (epsilon[:, :-1] + epsilon[:, 1:]) / 2
    eps_e[:, 1:] = (epsilon[:, :-1] + epsilon[:, 1:]) / 2
    # North/south half-points
    eps_n[:-1, :] = (epsilon[:-1, :] + epsilon[1:, :]) / 2
    eps_s[1:, :] = (epsilon[:-1, :] + epsilon[1:, :]) / 2

    # Build the finite difference matrix
    for i in range(N):
        for j in range(M):
            row = idx(i, j)

            # Coefficients for the 5-point stencil
            center = (
                -(eps_w[i, j] + eps_e[i, j]) / (dy**2 * fy)
                - (eps_n[i, j] + eps_s[i, j]) / (dx**2 * fx)
                + porous_source[i, j]
            )
            west = eps_w[i, j] / (dy**2 * fy)
            east = eps_e[i, j] / (dy**2 * fy)
            north = eps_n[i, j] / (dx**2 * fx)
            south = eps_s[i, j] / (dx**2 * fx)

            # center point
            A[row, row] = center

            # West neighbor
            if j > 0:
                A[row, idx(i, j - 1)] = west
            else:
                match bc["west"]:
                    case "Periodic":
                        A[row, idx(i, M - 1)] = west
                    case "Neumann":
                        A[row, row] += west
                    case "Dirichlet":
                        A[row, :] = 0
                        A[row, row] = 1
                        b[row] = bc_vals["west"] ** 2
                    case _:
                        raise ValueError("invalid BC")

            # East neighbor
            if j < M - 1:
                A[row, idx(i, j + 1)] = east
            else:
                match bc["east"]:
                    case "Periodic":
                        A[row, idx(i, 0)] = east
                    case "Neumann":
                        A[row, row] += east
                    case "Dirichlet":
                        A[row, :] = 0
                        A[row, row] = 1
                        b[row] = bc_vals["east"] ** 2
                    case _:
                        raise ValueError("invalid BC")

            # North neighbor
            if i > 0:
                A[row, idx(i - 1, j)] = north
            else:
                match bc["north"]:
                    case "Periodic":
                        A[row, idx(N - 1, j)] = north
                    case "Neumann":
                        A[row, row] += north
                    case "Dirichlet":
                        A[row, :] = 0
                        A[row, row] = 1
                        b[row] = bc_vals["north"] ** 2
                    case _:
                        raise ValueError("invalid BC")

            # South neighbor
            if i < N - 1:
                A[row, idx(i + 1, j)] = south
            else:
                match bc["south"]:
                    case "Periodic":
                        A[row, idx(0, j)] = south
                    case "Neumann":
                        A[row, row] += south
                    case "Dirichlet":
                        A[row, :] = 0
                        A[row, row] = 1
                        b[row] = bc_vals["south"] ** 2
                    case _:
                        raise ValueError("invalid BC")
            # Source term
            # b[row] += porous_source[i, j] * ps**2

    # Solve the linear system
    A = A.tocsr()

    p_flat = spla.spsolve(A, b) ** 0.5
    p = p_flat.reshape((N, M)).T
    return p



import numpy as np
from dataclasses import dataclass


@dataclass
class Result:
    """Class to hold bearing calculation results"""

    name: str
    p: np.ndarray
    w: np.ndarray
    k: np.ndarray
    qs: np.ndarray
    qa: np.ndarray
    qc: np.ndarray


# def get_y(bearing):
#     b = bearing
#     if b.ny == 1:
#         return 1
#     elif b.case =="rectangular":
#         return np.linspace(-b.ya / 2, b.ya / 2, b.ny)
#     elif b.case == "circular" or "annular":
#         return np.linspace(0, b.ya, b.ny)


def get_area(bearing):
    b = bearing
    match b.case:
        case "circular":
            A = np.pi * b.xa**2
        case "annular":
            A = np.pi * (b.xa**2 - b.xc**2)
        case "infinite":
            A = b.xa
        case "rectangular":
            A = b.xa * b.ya
        case "journal":
            A = 2 * np.pi * b.xa * b.ya
        case _:
            raise ValueError(f"Unknown case: {b.case}")

    return A


def get_geom(bearing):
    """
    Calculate the geometry of the bearing.
    """
    b = bearing
    if b.ny == 1:
        match b.error_type:
            case "none":
                geom = np.zeros(b.nx)
            case "linear":
                geom = b.error * (1 - b.x / b.xa)
            case "quadratic":
                geom = b.error * (1 - b.x**2 / b.xa**2)
            case _:
                raise ValueError(f"Unknown error type: {b.error_type}")

    else:
        if b.csys == "cartesian":

            x = b.x[:, None]
            y = b.y[None, :]
            zeros = np.zeros((b.nx, b.ny))
            match b.error_type:
                case "none":
                    geom = zeros
                case "linear":
                    geom = (
                        b.error
                        * 2
                        * np.maximum(np.abs(x) / b.xa + zeros, np.abs(y) / b.ya + zeros)
                    )
                case "quadratic":
                    geom = (
                        b.error
                        * 4
                        * np.maximum((x / b.xa) ** 2 + zeros, (y / b.ya) ** 2 + zeros)
                    )
                case "tiltx":
                    geom = b.error * b.x[:, None] / b.xa + zeros
                case "tilty":
                    geom = b.error * b.y / b.ya + zeros
                case _:
                    raise ValueError(f"Unknown error type: {b.error_type}")

        elif b.csys == "polar":
            r = b.x[:, None]
            theta = b.y[None, :]

            match b.error_type:
                case "none":
                    geom = np.zeros((b.nx, b.ny))
                case "linear":
                    geom = b.error * (1 - r / b.xa) + np.zeros_like(theta)
                case "quadratic":
                    geom = b.error * (1 - (r / b.xa) ** 2) + np.zeros_like(theta)
                case _:
                    raise ValueError(f"Unknown error type: {b.error_type}")

        else:
            raise ValueError(f"Unknown coordinate system: {b.csys}")

    return geom - np.min(geom)


def get_beta(bearing):
    """
    Calculate the porous feeding parameter.

    Returns:
        float: The porous feeding parameter, beta.
    """
    b = bearing
    beta = 6 * b.kappa * b.xa**2 / (b.hp * b.ha**3)
    return beta


def get_kappa(bearing):
    """
    Calculate the permeability.

    Returns:
        float: Permeability, kappa.
    """
    b = bearing

    if getattr(b, "blocked", False):
        kappa = (
            2 * b.Qsc / 6e4 * b.mu * b.hp * b.pa / (b.block_A * (b.psc**2 - b.pa**2))
        )
    else:
        kappa = 2 * b.Qsc / 6e4 * b.mu * b.hp * b.pa / (b.A * (b.psc**2 - b.pa**2))
    return round_to_sig_dig(kappa, 3)


def get_Qsc(bearing):
    """
    Calculate the permeability.

    Returns:
        float: Permeability, kappa.
    """
    b = bearing

    if b.blocked:
        Qsc = (
            b.kappa * 6e4 * b.block_A * (b.psc**2 - b.pa**2) / (2 * b.mu * b.hp * b.pa)
        )
    else:
        Qsc = b.kappa * 6e4 * b.A * (b.psc**2 - b.pa**2) / (2 * b.mu * b.hp * b.pa)
    return round_to_sig_dig(Qsc, 3)


def round_to_sig_dig(number, digits):
    return np.round(number, -int(np.floor(np.log10(np.abs(number)))) + (digits - 1))


def get_dA(bearing) -> np.ndarray:
    b = bearing
    if b.ny == 1:
        match b.csys:
            case "polar":
                dA = np.pi * np.gradient(b.x**2)
                dA[[0, -1]] = dA[[0, -1]] / 2
            case "cartesian":
                dA = b.dx.copy()
                dA[[0, -1]] = dA[[0, -1]] / 2
            case _:
                raise ValueError("Error: invalid csys in dA calculation")
    else:
        match b.csys:
            case "polar":
                dx2 = b.x**2 - np.insert(b.x[:-1], 0, bearing.xc) ** 2
                dy = b.y - np.insert(b.y[:-1], 0, 0)
                dA = 0.5 * dx2[None, :] * dy[:, None]
            case "cartesian":
                dA = b.dx * b.dy
            case _:
                raise ValueError("Error: invalid csys in dA calculation")
    return dA


def get_load_capacity(bearing, p: np.ndarray) -> np.ndarray:
    """
    Calculate the load capacity of the bearing.

    Args:
        p (numpy.ndarray): The pressure distribution.

    Returns:
        numpy.ndarray: The calculated load capacity.
    """
    b = bearing
    p_rel = p - b.pa
    if b.case == "journal":
        w_x = p_rel * np.cos(b.theta)[None, :, None] * b.xa * b.dx * b.dy
        w = np.sum(w_x, axis=(0, 1))
    elif p.ndim == 2:
        ny = b.ny
        b.ny = 1
        dA = get_dA(b)
        b.ny = ny
        w = np.sum(p_rel * dA[:, None], axis=0)
    else:
        dA = get_dA(b)
        w = np.sum(p_rel * dA, axis=(0, 1))
    return w


def get_stiffness(bearing, w):
    """
    Calculate the stiffness.

    Args:
        w (numpy.ndarray): The load or deflection values.

    Returns:
        numpy.ndarray: The stiffness values.
    """
    b = bearing
    if b.case == "journal":
        k = np.gradient(w, b.e.flatten())
    else:
        k = -np.gradient(w, b.ha.flatten())
    return k


def get_volumetric_flow(bearing, p: np.ndarray, soltype: str) -> tuple:
    """Calculate volumetric flow rates through the bearing.

    Args:
        bearing: Bearing instance containing geometry and properties
        p (np.ndarray): Pressure distribution array
        soltype (bool): Solution type (ANALYTIC or NUMERIC)

    Returns:
        tuple: (qs, qa, qc) where:
            - qs (np.ndarray): Supply flow rate (L/min)
            - qa (np.ndarray): Ambient flow rate (L/min)
            - qc (np.ndarray): Chamber flow rate (L/min)
    """
    b = bearing
    if soltype == "analytic" or soltype == "numeric":
        # 1D case
        match soltype:
            case "analytic":
                h = b.ha
            case "numeric":
                h = b.ha + b.geom[:, None]

        match b.csys:
            case "polar":
                q = (
                    -6e4
                    * h**3
                    * b.rho
                    * np.gradient(p**2, axis=0)
                    * np.pi
                    * b.x[:, None]
                    / (12 * b.mu * b.pa * b.dx[:, None])
                )
            case "cartesian":
                q = (
                    -6e4
                    * h**3
                    * b.rho
                    * np.gradient(p**2, axis=0)
                    / (12 * b.mu * b.pa * b.dx[:, None])
                )
            case _:
                raise ValueError("Invalid csys")

        qa = q[-1, :]
        qc = q[1, :]
        qs = qa - qc

    elif soltype == "numeric2d":
        h = b.ha[None, None, :] + b.geom.T[:, :, None]
        match b.csys:
            case "polar":
                qx = (
                    -6e4
                    * h**3
                    * b.rho
                    * np.gradient(p**2, axis=1)
                    / (2 * np.pi)
                    * b.x[None, :, None]
                    * b.dy[:, None, None]
                    / (12 * b.mu * b.pa * b.dx[None, :, None])
                )
                qy = 0
                qa = np.sum(qx[:, -1, :], axis=0)
                qc = np.sum(qx[:, 0, :], axis=0)
            case "cartesian":
                if b.case == "rectangular":
                    qx = (
                        -6e4
                        * h**3
                        * b.rho
                        * np.gradient(p**2, axis=1)
                        * b.dy
                        / (12 * b.mu * b.pa * b.dx)
                    )
                    qy = (
                        -6e4
                        * h**3
                        * b.rho
                        * np.gradient(p**2, axis=0)
                        * b.dx
                        / (12 * b.mu * b.pa * b.dy)
                    )

                    qa = np.sum(np.abs(qx[:, (0, -1), :]), axis=(0, 1)) + np.sum(
                        abs(qy[(0, -1), :, :]), axis=(0, 1)
                    )
                    qc = 0
                elif b.case == "journal":
                    h = np.transpose(b.clearance + b.geom[:, :, None], (1, 0, 2))
                    qx = (
                        -6e4
                        * h**3
                        * b.rho
                        * np.gradient(p**2, axis=1)
                        * b.dy
                        / (12 * b.mu * b.pa * b.dx)
                    )
                    qy = (
                        -6e4
                        * h**3
                        * b.rho
                        * np.gradient(p**2, axis=0)
                        * b.dx
                        * b.xa
                        / (12 * b.mu * b.pa * b.dy)
                    )
                    qa = np.sum(qy[-1, :, :], axis=(0))
                    qc = np.sum(qy[0, :, :], axis=(0))
            case _:
                qa, qc = 0, 0

        qs = qa - qc
    return qs, qa, qc



import pytest
from openairbearing.app.callbacks import get_bearing
from openairbearing.bearings import (
    CircularBearing,
    AnnularBearing,
    InfiniteLinearBearing,
    RectangularBearing,
)


def test_get_bearing():
    assert get_bearing("circular") == CircularBearing
    assert get_bearing("annular") == AnnularBearing
    assert get_bearing("infinite") == InfiniteLinearBearing
    assert get_bearing("rectangular") == RectangularBearing
    with pytest.raises(TypeError, match="no default bearing defined"):
        get_bearing("unknown")


# psutil multiprocess selenium "dash[testing]"
# def test_update_bearing(dash_duo):
#     app = import_app("openairbearing.app.app")  # Replace with the actual path to your Dash app
#     dash_duo.start_server(app)

#     # Simulate user selecting a bearing case
#     dash_duo.select_dcc_dropdown("#case-select", "circular")

#     # Simulate user input for parameters
#     dash_duo.find_element("#pa-input").send_keys("0.1")  # Ambient pressure in MPa
#     dash_duo.find_element("#ps-input").send_keys("0.6")  # Supply pressure in MPa

#     # Wait for the callback to update the outputs
#     dash_duo.wait_for_text_to_equal("#kappa-input", "calculated_value", timeout=5)
#     dash_duo.wait_for_text_to_equal("#Qsc-input", "calculated_value", timeout=5)

#     # Verify the output figures
#     assert dash_duo.find_element("#bearing-plots").is_displayed()
#     assert dash_duo.find_element("#bearing-shape").is_displayed()
